sub _myBizarroOperations_init { }

loadMacros('bizarroArithmetic.pl', 'parserRoot.pl');

################################################################################################
# Method to enable the bizarro operations for a given context.
################################################################################################

sub contextEnableBizarro {
	my ($context, %options) = @_;

	if ($options{enableAll}) {
		$options{enableSqrt}                 = 1;
		$options{enableExponents}            = 1;
		$options{enableRoots}                = 1;
		$options{enableAllBizarroArithmetic} = 1;
	}

	if ($options{enableAllBizarroArithmetic}) {
		$options{enableBizarroAdd}      = 1;
		$options{enableBizarroSubtract} = 1;
		$options{enableBizarroMultiply} = 1;
		$options{enableBizarroDivide}   = 1;
	}

	$context->functions->set(
		sqrt => {
			class             => 'my::Function::numeric',
			negativeIsComplex => $options{negativeIsComplex}
		}
	) if ($options{enableSqrt});

	$context->operators->set(
		'^' => {
			class             => 'my::Parser::BOP::power',
			isCommand         => 1,
			negativeIsComplex => $options{negativeIsComplex},
			perl              => 'my::Parser::BOP::power->_eval'
		},
		'**' => {
			class             => 'my::Parser::BOP::power',
			isCommand         => 1,
			negativeIsComplex => $options{negativeIsComplex},
			perl              => 'my::Parser::BOP::power->_eval'
		}
	) if ($options{enableExponents});

	parser::Root->Enable($context, $options{negativeIsComplex}),
		$context->functions->set(root => { class => 'my::Function::numeric2' })
		if ($options{enableRoots});

	$context->operators->set('+' => { class => 'bizarro::BOP::add', isCommand => 1 })
		if ($options{enableBizarroAdd});

	$context->operators->set('-' => { class => 'bizarro::BOP::subtract', isCommand => 1 })
		if ($options{enableBizarroSubtract});

	$context->operators->set(
		'*'  => { class => 'bizarro::BOP::multiply', isCommand => 1 },
		'* ' => { class => 'bizarro::BOP::multiply', isCommand => 1 },
		' *' => { class => 'bizarro::BOP::multiply', isCommand => 1 }
	) if ($options{enableBizarroMultiply});

	$context->operators->set(
		'/'  => { class => 'bizarro::BOP::divide', isCommand => 1 },
		'/ ' => { class => 'bizarro::BOP::divide', isCommand => 1 },
		' /' => { class => 'bizarro::BOP::divide', isCommand => 1 },
		'//' => { class => 'bizarro::BOP::divide', isCommand => 1 }
	) if ($options{enableBizarroDivide});

	$context->{format}{number} = '%.13f#';
	$context->flags->set(
		reduceConstantFunctions => 0,
		reduceConstants         => 0,
		formatStudentAnswer     => 'parsed',
		tolerance               => 1E-13,
		tolType                 => 'absolute'
	);

	return;
}

################################################################################################
# Extract formulas for the numbers (or inifinities) in the answer.  This takes an original
# formula, and returns an array of formulas for the numbers in the answer.  The structure of the
# returned array will be essentially the same as that of the original formula.
################################################################################################

sub extractFormulas {
	my $orig_formula = shift;
	return if (!Value::isFormula($orig_formula) || $orig_formula eq '');

	return [ Parser::Formula("$orig_formula") ]
		if Value::isNumber($orig_formula) || Value::matchInfinite($orig_formula);

	my @formula_list = ();
	my @entries      = (
		$orig_formula->type eq 'Union' ? $orig_formula->{tree}->makeUnion
		: (($orig_formula->type eq 'Assignment' || ref($orig_formula->{tree}) eq 'Parser::BOP::equality')
			? ($orig_formula->{tree}{lop}, $orig_formula->{tree}{rop})
			: @{ $orig_formula->{tree}{coords} })
	);
	for my $entry (@entries) {
		my $v = Parser::Formula($entry);
		if (Value::isNumber($v) || Value::matchInfinite($v)) {
			push(@formula_list, Parser::Formula("$v"));
			next;
		}
		push(@formula_list, extractFormulas($v));
	}

	return [@formula_list];
}

################################################################################################
# Compare a student formula to correct formulas.
#
# $cor is a reference to an array of correct formulas to compare to, $stu a formula to check,
# and %flags the bizarro flags to set for the comparison.
#
# Returns 1 if the student formula is equal to one of the correct formulas with the bizarro
# flags passed in, and 0 otherwise.
################################################################################################

sub bizarroCmp {
	my ($cor, $stu, %flags) = @_;
	my $ret = 0;

	Context()->flags->set(%flags);
	my $recompute = Context()->flags->get('forceRecompute');
	my $stu_f     = $recompute ? Formula("$stu") : $stu;
	for (@$cor) {
		my $cor_f = $recompute ? Formula("$_") : $_;
		delete $stu_f->{test_values};
		delete $cor_f->{test_values};
		$ret = 1 if ($cor_f == $stu_f);
	}
	for (keys(%flags)) { $flags{$_} = 0; }
	Context()->flags->set(%flags);

	return $ret;
}

################################################################################################
# Override square roots and exponents to return special values when evaluated in order to check
# for a lack of simplification.
################################################################################################

package my::Function::numeric;
our @ISA = qw(Parser::Function::numeric);

sub sqrt {
	my ($self, @inputs) = @_;

	my $value = $self->context->flag('setSqrt') // 0;
	return $value * $inputs[0] + 0.5 if $value;

	if ($inputs[0] < 0 && $self->{def}{negativeIsComplex}) {
		my $x = Value::makeValue($inputs[0], context => $self->context);
		return $self->Package('Complex')->promote($self->context, $x)->sqrt;
	}
	return $self->SUPER::sqrt(@inputs);
}

package my::Parser::BOP::power;
our @ISA = qw(Parser::BOP::power);

sub _eval {
	my ($self, @inputs) = @_;

	my $value = (
		($inputs[1] == 0.5 && $self->context->flag('setSqrt')) ? $self->context->flag('setSqrt')
		: (($inputs[1] == 2 && $self->context->flag('setSquare')) ? $self->context->flag('setSquare')
			: ($self->context->flag('setExponent') // 0))
	);
	return $value * $inputs[0] + $inputs[1] if $value;

	if ($inputs[0] < 0 && $self->{def}{negativeIsComplex} && $inputs[1] != int($inputs[1])) {
		my $a = Value::makeValue($inputs[0], context => $self->context);
		return $self->Package('Complex')->promote($self->context, $a)**$inputs[1];
	}
	return $self->SUPER::_eval(@inputs);
}

package my::Function::numeric2;
our @ISA = ('parser::Root::Function::numeric2');

sub root {
	my ($self, @inputs) = @_;

	my $value = $self->context->flag('setRoot') // 0;
	return 0                                    if $value && $inputs[0] == 0;
	return $value * $inputs[1] + 1 / $inputs[0] if $value;

	return $self->SUPER::root(@inputs);
}

1;
